@@ -3,9 +3,9 @@ use futures::{
     Future,
     task::{Context, Poll}
 };
-use mio::unix::UnixReady;
 pub use mio_aio::LioError;
-use tokio::io::PollEvented;
+use nix::errno::Errno;
+use tokio::io::bsd::{AioSource, Aio};
 use std::{fs, io, mem};
 use std::os::unix::fs::FileTypeExt;
 use std::os::unix::io::{AsRawFd, RawFd};
@@ -26,12 +26,68 @@ nix::ioctl_read! {
     diocgstripesize, 'd', 139, nix::libc::off_t
 }
 
+fn conv_poll_err<T>(e: io::Error) -> Poll<Result<T, nix::Error>> {
+    let raw = e.raw_os_error().unwrap_or(0);
+    let errno = Errno::from_i32(raw);
+    Poll::Ready(Err(errno))
+}
+
+macro_rules! lio_resubmit {
+    ($self: ident, $ev: expr) => {
+        {
+            // Some requests must've completed; now issue the rest.
+            let result = (*$self.op.as_mut().unwrap()).0.resubmit();
+            $self.op
+                .as_mut()
+                .unwrap()
+                .clear_ready($ev);
+            match result {
+                Ok(()) => {
+                    $self.state = AioState::InProgress;
+                    None
+                },
+                Err(LioError::EINCOMPLETE) => None,
+                Err(LioError::EAGAIN) => Some(Err(Errno::EAGAIN)),
+                Err(LioError::EIO(_)) => Some(Err(Errno::EIO)),
+            }
+        }
+    }
+}
+
+/// Must wrap the real mio::event::Source in order to implement AioSource, which
+/// exists because Tokio has a policy of not exposing any mio types in its
+/// public API.
+#[derive(Debug)]
+struct WrappedAioCb<'a>(mio_aio::AioCb<'a>);
+impl<'a> AioSource for WrappedAioCb<'a> {
+    fn register(&mut self, kq: RawFd, token: usize) {
+        self.0.register_raw(kq, token)
+    }
+    fn deregister(&mut self) {
+        self.0.deregister_raw()
+    }
+}
+
+/// Must wrap the real mio::event::Source in order to implement AioSource, which
+/// exists because Tokio has a policy of not exposing any mio types in its
+/// public API.
+#[derive(Debug)]
+struct WrappedLioCb<'a>(mio_aio::LioCb<'a>);
+impl<'a> AioSource for WrappedLioCb<'a> {
+    fn register(&mut self, kq: RawFd, token: usize) {
+        self.0.register_raw(kq, token)
+    }
+    fn deregister(&mut self) {
+        self.0.deregister_raw()
+    }
+}
+
 // LCOV_EXCL_START
 #[derive(Debug)]
 enum AioOp<'a> {
-    Fsync(PollEvented<mio_aio::AioCb<'static>>),
-    Read(PollEvented<mio_aio::AioCb<'a>>),
-    Write(PollEvented<mio_aio::AioCb<'a>>),
+    Fsync(Aio<WrappedAioCb<'static>>),
+    Read(Aio<WrappedAioCb<'a>>),
+    Write(Aio<WrappedAioCb<'a>>),
 }
 // LCOV_EXCL_STOP
 
@@ -67,9 +123,9 @@ impl<'a> AioFut<'a> {
     fn aio_return(&mut self) -> Result<Option<isize>, nix::Error> {
         match self.op {
             AioOp::Fsync(ref mut io) =>
-                io.get_mut().aio_return().map(|_| None),
+                (*io).0.aio_return().map(|_| None),
             AioOp::Read(ref mut io) | AioOp::Write(ref mut io) =>
-                io.get_mut().aio_return().map(Some),
+                (*io).0.aio_return().map(Some),
         }
     }
 }
@@ -82,25 +138,25 @@ pub struct AioResult {
     pub value: Option<isize>,
 }
 
-/// The return value of [`readv_at`]
+/// The return value of [`File::readv_at`]
 #[must_use = "futures do nothing unless polled"]
 #[allow(clippy::type_complexity)]
 pub struct ReadvAt<'a> {
-    op: Option<PollEvented<mio_aio::LioCb<'a>>>,
-    /// If needed, bufsav.0 combines [`readv_at`]'s argument slices into a
+    op: Option<Aio<WrappedLioCb<'a>>>,
+    /// If needed, bufsav.0 combines [`File::readv_at`]'s argument slices into a
     /// bigger slice that satisfies sectorsize requirements.  After completion,
     /// the data will be copied back to bufsav.1
     bufsav: Option<(Pin<Box<[u8]>>, &'a mut [&'a mut [u8]])>,
     state: AioState,
 }
 
-/// The return value of [`writev_at`]
+/// The return value of [`File::writev_at`]
 #[must_use = "futures do nothing unless polled"]
 pub struct WritevAt<'a> {
-    op: Option<PollEvented<mio_aio::LioCb<'a>>>,
-    /// If needed, _accumulator combines [`writev_at`]'s argument slices into a
-    /// bigger slice that satisfies sectorsize requirements, and owns the data
-    /// for the lifetime of [`op`].
+    op: Option<Aio<WrappedLioCb<'a>>>,
+    /// If needed, _accumulator combines [`File::writev_at`]'s argument slices
+    /// into a bigger slice that satisfies sectorsize requirements, and owns the
+    /// data for the lifetime of [`WritevAt::op`].
     _accumulator: Option<Pin<Box<[u8]>>>,
     state: AioState,
 }
@@ -109,12 +165,8 @@ impl<'a> Future for ReadvAt<'a> {
     type Output = Result<usize, nix::Error>;
 
     fn poll(mut self: Pin<&mut Self>, cx: &mut Context) -> Poll<Self::Output> {
-        let poll_result = self.op
-                              .as_mut()
-                              .unwrap()
-                              .poll_read_ready(cx, UnixReady::lio().into());
         if let AioState::Allocated = self.state {
-            let result = self.op.as_mut().unwrap().get_mut().submit();
+            let result = (*self.op.as_mut().unwrap()).0.submit();
             match result {
                 Ok(()) => self.state = AioState::InProgress,
                 Err(LioError::EINCOMPLETE) => {
@@ -123,77 +175,68 @@ impl<'a> Future for ReadvAt<'a> {
                     self.state = AioState::Incomplete;
                 },
                 Err(LioError::EAGAIN) =>
-                    return Poll::Ready(Err(nix::Error::Sys(nix::errno::Errno::EAGAIN))),
+                    return Poll::Ready(Err(Errno::EAGAIN)),
                 Err(LioError::EIO(_)) => {
-                    return Poll::Ready(Err(nix::Error::Sys(nix::errno::Errno::EIO)))
+                    return Poll::Ready(Err(Errno::EIO))
                 },
             }
         }
-        if !poll_result.is_ready() {
-            return Poll::Pending;
-        }
-        if AioState::Incomplete == self.state {
-            // Some requests must've completed; now issue the rest.
-            let result = self.op.as_mut().unwrap().get_mut().resubmit();
-            self.op
-                .as_mut()
-                .unwrap()
-                .clear_read_ready(cx, UnixReady::lio().into())
-                .unwrap();
-            match result {
-                Ok(()) => {
-                    self.state = AioState::InProgress;
-                    return Poll::Pending;
-                },
-                Err(LioError::EINCOMPLETE) => {
-                    return Poll::Pending;
-                },
-                Err(LioError::EAGAIN) =>
-                    return Poll::Ready(Err(nix::Error::Sys(nix::errno::Errno::EAGAIN))),
-                Err(LioError::EIO(_)) =>
-                    return Poll::Ready(Err(nix::Error::Sys(nix::errno::Errno::EIO))),
-            }
-        }
-        let r = self.op.take()
-            .unwrap()
-            .into_inner().unwrap()
-            .into_results(|mut iter|
-                iter.try_fold(0, |total, lr|
-                    lr.result.map(|r| total + r as usize)
-                )
-            );
-        if let Ok(v) = r {
-            if let Some((accum, ob)) = &mut self.bufsav  {
-                // Copy results back into the individual buffers
-                let mut i = 0;
-                let mut j = 0;
-                let mut total = 0;
-                while total < v {
-                    let z = (v - total).min(ob[i].len() - j);
-                    ob[i][j..j + z].copy_from_slice(&accum[total..total + z]);
-                    j += z;
-                    total += z;
-                    if j == ob[i].len() {
-                        j = 0;
-                        i += 1;
+        loop {
+            let poll_result = self.op
+                                  .as_mut()
+                                  .unwrap()
+                                  .poll_ready(cx);
+            match poll_result {
+                Poll::Pending => break Poll::Pending,
+                Poll::Ready(Err(e)) => break conv_poll_err(e),
+                Poll::Ready(Ok(ev)) => {
+                    if AioState::Incomplete == self.state {
+                        if let Some(r) = lio_resubmit!(self, ev) {
+                            break Poll::Ready(r);
+                        }
+                    } else {
+                        let r = self.op.take()
+                            .unwrap()
+                            .into_inner()
+                            .0
+                            .into_results(|mut iter|
+                                iter.try_fold(0, |total, lr|
+                                    lr.result.map(|r| total + r as usize)
+                                )
+                            );
+                        if let Ok(v) = r {
+                            if let Some((accum, ob)) = &mut self.bufsav  {
+                                // Copy results back into the individual buffers
+                                let mut i = 0;
+                                let mut j = 0;
+                                let mut tot = 0;
+                                while tot < v {
+                                    let z = (v - tot).min(ob[i].len() - j);
+                                    ob[i][j..j + z]
+                                        .copy_from_slice(&accum[tot..tot + z]);
+                                    j += z;
+                                    tot += z;
+                                    if j == ob[i].len() {
+                                        j = 0;
+                                        i += 1;
+                                    }
+                                }
+                            }
+                        }
+                        break Poll::Ready(r)
                     }
                 }
             }
         }
-        Poll::Ready(r)
-}
+    }
 }
 
 impl<'a> Future for WritevAt<'a> {
     type Output = Result<usize, nix::Error>;
 
     fn poll(mut self: Pin<&mut Self>, cx: &mut Context) -> Poll<Self::Output> {
-        let poll_result = self.op
-                              .as_mut()
-                              .unwrap()
-                              .poll_read_ready(cx, UnixReady::lio().into());
         if let AioState::Allocated = self.state {
-            let result = self.op.as_mut().unwrap().get_mut().submit();
+            let result = (*self.op.as_mut().unwrap()).0.submit();
             match result {
                 Ok(()) => self.state = AioState::InProgress,
                 Err(LioError::EINCOMPLETE) => {
@@ -202,46 +245,40 @@ impl<'a> Future for WritevAt<'a> {
                     self.state = AioState::Incomplete;
                 },
                 Err(LioError::EAGAIN) =>
-                    return Poll::Ready(Err(nix::Error::Sys(nix::errno::Errno::EAGAIN))),
+                    return Poll::Ready(Err(Errno::EAGAIN)),
                 Err(LioError::EIO(_)) => {
-                    return Poll::Ready(Err(nix::Error::Sys(nix::errno::Errno::EIO)))
+                    return Poll::Ready(Err(Errno::EIO))
                 },
             }
         }
-        if !poll_result.is_ready() {
-            return Poll::Pending;
-        }
-        if AioState::Incomplete == self.state {
-            // Some requests must've completed; now issue the rest.
-            let result = self.op.as_mut().unwrap().get_mut().resubmit();
-            self.op
-                .as_mut()
-                .unwrap()
-                .clear_read_ready(cx, UnixReady::lio().into())
-                .unwrap();
-            match result {
-                Ok(()) => {
-                    self.state = AioState::InProgress;
-                    return Poll::Pending;
-                },
-                Err(LioError::EINCOMPLETE) => {
-                    return Poll::Pending;
-                },
-                Err(LioError::EAGAIN) =>
-                    return Poll::Ready(Err(nix::Error::Sys(nix::errno::Errno::EAGAIN))),
-                Err(LioError::EIO(_)) =>
-                    return Poll::Ready(Err(nix::Error::Sys(nix::errno::Errno::EIO))),
+        loop {
+            let poll_result = self.op
+                                  .as_mut()
+                                  .unwrap()
+                                  .poll_ready(cx);
+            match poll_result {
+                Poll::Pending => break Poll::Pending,
+                Poll::Ready(Err(e)) => break conv_poll_err(e),
+                Poll::Ready(Ok(ev)) => {
+                    if AioState::Incomplete == self.state {
+                        if let Some(r) = lio_resubmit!(self, ev) {
+                            break Poll::Ready(r);
+                        }
+                    } else {
+                        let r = self.op.take()
+                            .unwrap()
+                            .into_inner()
+                            .0
+                            .into_results(|mut iter|
+                                iter.try_fold(0, |total, lr|
+                                    lr.result.map(|r| total + r as usize)
+                                )
+                            );
+                        break Poll::Ready(r);
+                    }
+                }
             }
         }
-        let r = self.op.take()
-            .unwrap()
-            .into_inner().unwrap()
-            .into_results(|mut iter|
-                iter.try_fold(0, |total, lr|
-                    lr.result.map(|r| total + r as usize)
-                )
-            );
-        Poll::Ready(r)
     }
 }
 
@@ -345,7 +382,7 @@ impl File {
     /// use std::fs;
     /// use std::io::Write;
     /// use tempfile::TempDir;
-    /// use tokio::runtime::Runtime;
+    /// use tokio::runtime;
     ///
     /// const WBUF: &[u8] = b"abcdef";
     /// const EXPECT: &[u8] = b"cdef";
@@ -360,7 +397,10 @@ impl File {
     ///     .open(&path)
     ///     .map(tokio_file::File::new)
     ///     .unwrap();
-    /// let mut rt = Runtime::new().unwrap();
+    /// let rt = runtime::Builder::new_current_thread()
+    ///     .enable_io()
+    ///     .build()
+    ///     .unwrap();
     /// let r = rt.block_on(async {
     ///     file.read_at(&mut rbuf[..], 2).unwrap().await
     /// }).unwrap();
@@ -374,7 +414,8 @@ impl File {
                             buf,
                             0,  //priority
                             mio_aio::LioOpcode::LIO_NOP);
-        PollEvented::new(aiocb)
+        let source = WrappedAioCb(aiocb);
+        Aio::new_for_aio(source)
         .map(|pe| AioFut {
             op: AioOp::Read(pe),
             state: AioState::Allocated
@@ -413,7 +454,7 @@ impl File {
     /// use std::fs;
     /// use std::io::Write;
     /// use tempfile::TempDir;
-    /// use tokio::runtime::Runtime;
+    /// use tokio::runtime;
     ///
     /// const WBUF: &[u8] = b"abcdefghijklmnopqrwtuvwxyz";
     /// const EXPECT0: &[u8] = b"cdef";
@@ -434,7 +475,10 @@ impl File {
     ///     .open(&path)
     ///     .map(tokio_file::File::new)
     ///     .unwrap();
-    /// let mut rt = Runtime::new().unwrap();
+    /// let rt = runtime::Builder::new_current_thread()
+    ///     .enable_io()
+    ///     .build()
+    ///     .unwrap();
     /// let mut r = rt.block_on(async {
     ///     file.readv_at(&mut rbufs[..], 2).unwrap().await
     /// }).unwrap();
@@ -487,7 +531,8 @@ impl File {
             }
         }
         let liocb = builder.finish();
-        PollEvented::new(liocb)
+        let source = WrappedLioCb(liocb);
+        Aio::new_for_lio(source)
         .map(|pe| ReadvAt {
             op: Some(pe),
             bufsav,
@@ -505,7 +550,7 @@ impl File {
     /// use std::fs;
     /// use std::io::Read;
     /// use tempfile::TempDir;
-    /// use tokio::runtime::Runtime;
+    /// use tokio::runtime;
     ///
     /// let contents = b"abcdef";
     /// let mut rbuf = Vec::new();
@@ -518,7 +563,10 @@ impl File {
     ///     .open(&path)
     ///     .map(tokio_file::File::new)
     ///     .unwrap();
-    /// let mut rt = Runtime::new().unwrap();
+    /// let rt = runtime::Builder::new_current_thread()
+    ///     .enable_io()
+    ///     .build()
+    ///     .unwrap();
     /// let r = rt.block_on(async {
     ///     file.write_at(contents, 0).unwrap().await
     /// }).unwrap();
@@ -535,7 +583,8 @@ impl File {
         let fd = self.file.as_raw_fd();
         let aiocb = mio_aio::AioCb::from_slice(fd, offset, buf, 0,
             mio_aio::LioOpcode::LIO_NOP);
-        PollEvented::new(aiocb)
+        let source = WrappedAioCb(aiocb);
+        Aio::new_for_aio(source)
         .map(|pe| AioFut{
             op: AioOp::Write(pe),
             state: AioState::Allocated
@@ -572,7 +621,7 @@ impl File {
     /// use std::fs;
     /// use std::io::Read;
     /// use tempfile::TempDir;
-    /// use tokio::runtime::Runtime;
+    /// use tokio::runtime;
     ///
     /// const EXPECT: &[u8] = b"abcdefghij";
     /// let wbuf0 = b"abcdef";
@@ -588,7 +637,10 @@ impl File {
     ///     .open(&path)
     ///     .map(tokio_file::File::new)
     ///     .unwrap();
-    /// let mut rt = Runtime::new().unwrap();
+    /// let rt = runtime::Builder::new_current_thread()
+    ///     .enable_io()
+    ///     .build()
+    ///     .unwrap();
     /// let r = rt.block_on(async {
     ///     file.writev_at(&wbufs[..], 0).unwrap().await
     /// }).unwrap();
@@ -643,7 +695,8 @@ impl File {
             }
         }
         let liocb = builder.finish();
-        PollEvented::new(liocb)
+        let source = WrappedLioCb(liocb);
+        Aio::new_for_lio(source)
         .map(|pe| 
              WritevAt {
                 _accumulator: accumulator,
@@ -662,7 +715,7 @@ impl File {
     /// use std::fs;
     /// use std::io::Write;
     /// use tempfile::TempDir;
-    /// use tokio::runtime::Runtime;
+    /// use tokio::runtime;
     ///
     /// let dir = TempDir::new().unwrap();
     /// let path = dir.path().join("foo");
@@ -673,7 +726,10 @@ impl File {
     ///     .open(&path)
     ///     .map(tokio_file::File::new)
     ///     .unwrap();
-    /// let mut rt = Runtime::new().unwrap();
+    /// let rt = runtime::Builder::new_current_thread()
+    ///     .enable_io()
+    ///     .build()
+    ///     .unwrap();
     /// let r = rt.block_on(async {
     ///     file.sync_all().unwrap().await
     /// }).unwrap();
@@ -683,7 +739,8 @@ impl File {
         let aiocb = mio_aio::AioCb::from_fd(self.file.as_raw_fd(),
                             0,  //priority
                             );
-        PollEvented::new(aiocb)
+        let source = WrappedAioCb(aiocb);
+        Aio::new_for_aio(source)
         .map(|pe| AioFut{
             op: AioOp::Fsync(pe),
             state: AioState::Allocated
@@ -703,31 +760,36 @@ impl<'a> Future for AioFut<'a> {
     fn poll(mut self: Pin<&mut Self>, cx: &mut Context) -> Poll<Self::Output> {
         let poll_result = match self.op {
                 AioOp::Fsync(ref mut io) =>
-                    io.poll_read_ready(cx, UnixReady::aio().into()),
+                    io.poll_ready(cx),
                 AioOp::Read(ref mut io) =>
-                    io.poll_read_ready(cx, UnixReady::aio().into()),
+                    io.poll_ready(cx),
                 AioOp::Write(ref mut io) =>
-                    io.poll_read_ready(cx, UnixReady::aio().into()),
+                    io.poll_ready(cx),
         };
-        if !poll_result.is_ready() {
-            if self.state == AioState::Allocated {
-                let r = match self.op {
-                    AioOp::Fsync(ref mut pe) => pe.get_mut()
-                        .fsync(mio_aio::AioFsyncMode::O_SYNC),
-                    AioOp::Read(ref mut pe) => pe.get_mut().read(),
-                    AioOp::Write(ref mut pe) => pe.get_mut().write(),
-                };
-                if let Err(e) = r {
-                    return Poll::Ready(Err(e));
+        match poll_result {
+            Poll::Pending => {
+                if self.state == AioState::Allocated {
+                    let r = match self.op {
+                        AioOp::Fsync(ref mut pe) => (*pe).0
+                            .fsync(mio_aio::AioFsyncMode::O_SYNC),
+                        AioOp::Read(ref mut pe) => (*pe).0.read(),
+                        AioOp::Write(ref mut pe) => (*pe).0.write(),
+                    };
+                    if let Err(e) = r {
+                        return Poll::Ready(Err(e));
+                    }
+                    self.state = AioState::InProgress;
+                }
+                Poll::Pending
+            },
+            Poll::Ready(Err(e)) => conv_poll_err(e),
+            Poll::Ready(Ok(_ev)) => {
+                let result = self.aio_return();
+                match result {
+                    Ok(x) => Poll::Ready(Ok(AioResult{value: x})),
+                    Err(x) => Poll::Ready(Err(x))
                 }
-                self.state = AioState::InProgress;
             }
-            return Poll::Pending;
-        }
-        let result = self.aio_return();
-        match result {
-            Ok(x) => Poll::Ready(Ok(AioResult{value: x})),
-            Err(x) => Poll::Ready(Err(x))
         }
     }
 }
